"""Functions for reading DICOM files ith dicomity"""

from typing import Optional

import numpy as np
from pydicom import filereader, FileDataset

from dicomity.dictionary import DicomDictionary
from dicomity.types import GroupingMetadata


def read_dicom_image(file_name: str) -> np.array:
    meta = filereader.dcmread(file_name, stop_before_pixels=False)
    return meta.pixel_array


def read_dicom_tags(
        file_name: str,
        dictionary: Optional[DicomDictionary] = None,
        specific_tags: Optional[list[str]] = None
) -> FileDataset:
    if dictionary:
        specific_tags = dictionary.allTags()
    meta = filereader.dcmread(
        file_name,
        stop_before_pixels=True,
        specific_tags=specific_tags
    )
    return meta


def read_grouping_metadata(file_name: str) -> GroupingMetadata:
    meta = filereader.dcmread(fp=file_name,
                              stop_before_pixels=True,
                              specific_tags=GroupingMetadata.groupingTags())
    return GroupingMetadata.fromPyDicom(meta)


def is_dicom(file_name: str) -> bool:
    """Test whether a file is in DICOM format

    Uses a simple test for the DICM preamble - this is very fast but
    may occasionally produce false positives.

    For a more robust guarantee that a file really is DICOM, first call this
    function and if True, try to parse the file - if the tag data cannot be
    parsed then it is likely not a DICOM file

    Args:
        file_name: path and filename of the file to test

    Returns:
        True if the file appears to be a Dicom file
    """

    with open(file_name, "rb") as f:
        f.seek(128)
        preamble = f.read(4)
        return preamble == b"DICM"


def is_dicom_image_file(file_name: str) -> bool:
    """Tests if a file is a DICOM file but not a DICOMDIR file

    Args:
        file_name: Full path to file

    Returns:
        True if the file is DICOM and not a DICOMDIR
    """
    return file_name != 'DICOMDIR' and is_dicom(file_name)



import numpy as np

from dicomity.group import DicomGrouper
from dicomity.group import DicomStack


def test_grouper():
    grouper = DicomGrouper()
    assert grouper.largest_stack() == None
    # grouper.add_item()


def test_sort_images_by_location():
    group = DicomStack()
    group.add(filename="1", metadata=meta({'ImagePositionPatient': [0, 0, 1], 'ImageOrientationPatient': [1, 0, 0, 0, 1, 0]}))

    # grouping.append()
    sorted_indices, slice_thickness, global_origin_mm, sorted_positions = group.sort_images_by_location()
    assert np.array_equal(sorted_indices, [0])
    assert np.array_equal(sorted_positions, [[0, 0, 1]])
    assert slice_thickness == None
    assert global_origin_mm == [0, 0, 1]

    group.add(filename="2", metadata=meta({'ImagePositionPatient': [0, 0, 0], 'ImageOrientationPatient': [1, 0, 0, 0, 1, 0]}))
    sorted_indices, slice_thickness, global_origin_mm, sorted_positions = group.sort_images_by_location()
    assert np.array_equal(sorted_indices, [1, 0])
    assert np.array_equal(sorted_positions, [[0, 0, 0], [0, 0, 1]])
    assert slice_thickness == 1.0
    assert global_origin_mm == [0, 0, 0]

    group.add(filename="3", metadata=meta({'ImagePositionPatient': [0, 0, 2], 'ImageOrientationPatient': [1, 0, 0, 0, 1, 0]}))
    sorted_indices, slice_thickness, global_origin_mm, sorted_positions = group.sort_images_by_location()
    assert np.array_equal(sorted_indices, [1, 0, 2])
    assert np.array_equal(sorted_positions, [[0, 0, 0], [0, 0, 1], [0, 0, 2]])
    assert slice_thickness == 1.0
    assert global_origin_mm == [0, 0, 0]


def test_sort_images_by_instance():
    group = DicomStack()
    group.add(filename="1", metadata=meta({'InstanceNumber': 2}))
    sorted_indices, slice_thickness, global_origin_mm, sorted_positions = group.sort_images_by_location()
    assert np.array_equal(sorted_indices, [0])
    group.add(filename="2", metadata=meta({'InstanceNumber': 3}))
    sorted_indices, slice_thickness, global_origin_mm, sorted_positions = group.sort_images_by_location()
    assert np.array_equal(sorted_indices, [0, 1])
    group.add(filename="3", metadata=meta({'InstanceNumber': 1}))
    sorted_indices, slice_thickness, global_origin_mm, sorted_positions = group.sort_images_by_location()
    assert np.array_equal(sorted_indices, [2, 0, 1])


class DummyMeta:
    pass


def meta(din: dict):
    d = DummyMeta()
    for key, value in din.items():
        setattr(d, key, value)
    return d



from dicomity.types import TagDefinition


class DicomDictionary:
    """Build a dictionary for use with read_dicom_tags
        
    DicomDictionary can be used to construct a dictionary containing
    the Dicom tags you wish to read in using read_dicom_tags. A
    smaller subset of tags can reduce the time taken to parse a Dicom
    file.
        
    Usage:
        Several class methods provide useful dictionaries for most purposes:

         dictionary = DicomDictionary.grouping_dictionary()
             creates a dictionary of tags needed for grouping files
             together into volumes

         dictionary = DicomDictionary.essential_dictionary()
             creates a dictionary containing tags that are commonly used,
             including the pixel data

         dictionary = DicomDictionary.essential_dictionary_without_pixel_data()
             creates a dictionary containing tags that are commonly used,
             excluding the pixel data

         dictionary = DicomDictionary.complete_dictionary()
             creates a dictionary containing most public tags, including
             the pixel data
        
    """

    grouping_dictionary_without_pixel_data_singleton = None
    essential_dictionary_singleton = None
    essential_tags_dictionary_without_pixel_data_singleton = None
    complete_dictionary_singleton = None

    def __init__(self, tags):
        # The item delimiter is a special case. It must always be the
        # last tag in the list, so that the tag reading algorithm can
        # choose to ignore it
        tags.append(DicomDictionary._item_delimiter_tag())

        self.Dictionary = tags
        self._make_tag_list()
        self._make_tag_map()

    @classmethod
    def grouping_dictionary(cls):
        if not cls.grouping_dictionary_without_pixel_data_singleton:
            cls.grouping_dictionary_without_pixel_data_singleton = \
                DicomDictionary._create_grouping_tags_dictionary(
                    add_pixel_data=False)
        return cls.grouping_dictionary_without_pixel_data_singleton

    @classmethod
    def essential_dictionary(cls):
        if not cls.essential_dictionary_singleton:
            cls.essential_dictionary_singleton = \
                DicomDictionary._create_essential_tags_dictionary(
                    add_pixel_data=True)
        return cls.essential_dictionary_singleton

    @classmethod
    def essential_dictionary_without_pixel_data(cls):
        if not cls.essential_tags_dictionary_without_pixel_data_singleton:
            cls.essential_tags_dictionary_without_pixel_data_singleton \
                = DicomDictionary._create_essential_tags_dictionary(
                    add_pixel_data=False)
        return cls.essential_tags_dictionary_without_pixel_data_singleton

    @classmethod
    def complete_dictionary(cls):
        if not cls.complete_dictionary_singleton:
            cls.complete_dictionary_singleton = \
                DicomDictionary._create_all_tags_dictionary(add_pixel_data=True)
        return cls.complete_dictionary_singleton

    def _make_tag_list(self):
        self.tag_list = [d.TagIndex for d in self.Dictionary]

    def _make_tag_map(self):
        self.tag_map = {}
        for t in self.Dictionary:
            self.tag_map[t.TagIndex] = t

    @staticmethod
    def _create_grouping_tags_dictionary(add_pixel_data):
        tags = DicomDictionary._grouping_tags()
        if add_pixel_data:
            tags.append(DicomDictionary._pixel_data_tag())
        return DicomDictionary(tags)

    @staticmethod
    def _create_essential_tags_dictionary(add_pixel_data):
        tags = DicomDictionary._essentialTags()
        if add_pixel_data:
            tags.append(DicomDictionary._pixel_data_tag())
        return DicomDictionary(tags)

    @staticmethod
    def _create_all_tags_dictionary(add_pixel_data):
        tags = DicomDictionary.allTags()
        if add_pixel_data:
            tags.append(DicomDictionary._pixel_data_tag())
        return DicomDictionary(tags)

    @staticmethod
    def _pixel_data_tag():
        return TagDefinition('7FE0,0010', 'OB', 'PixelData')

    @staticmethod
    def _item_delimiter_tag():
        return TagDefinition('FFFE,E00D', 'UN', 'ItemDelimitationItem')

    @staticmethod
    def _grouping_tags():
        # pylint: disable=line-too-long
        tags = []
        tags.append(TagDefinition('0002,0000', 'UL', 'FileMetaInformationGroupLength'))
        tags.append(TagDefinition('0002,0010', 'UI', 'TransferSyntaxUID'))
        tags.append(TagDefinition('0008,0018', 'UI', 'SOPInstanceUID'))
        tags.append(TagDefinition('0008,0021', 'DA', 'SeriesDate'))
        tags.append(TagDefinition('0008,0031', 'TM', 'SeriesTime'))
        tags.append(TagDefinition('0008,0060', 'CS', 'Modality'))
        tags.append(TagDefinition('0008,1030', 'LO', 'StudyDescription'))
        tags.append(TagDefinition('0008,103E', 'LO', 'SeriesDescription'))
        tags.append(TagDefinition('0010,0010', 'PN', 'PatientName'))
        tags.append(TagDefinition('0010,0020', 'LO', 'PatientID'))
        tags.append(TagDefinition('0020,000D', 'UI', 'StudyInstanceUID'))
        tags.append(TagDefinition('0020,000E', 'UI', 'SeriesInstanceUID'))
        return tags

    @staticmethod
    def _essentialTags():
        # pylint: disable=line-too-long
        tags = []
        tags.append(TagDefinition('0002,0000', 'UL', 'FileMetaInformationGroupLength'))
        tags.append(TagDefinition('0002,0002', 'UI', 'MediaStorageSOPClassUID'))
        tags.append(TagDefinition('0002,0010', 'UI', 'TransferSyntaxUID'))
        tags.append(TagDefinition('0002,0012', 'UI', 'ImplementationClassUID'))

        tags.append(TagDefinition('0008,0008', 'CS', 'ImageType'))
        tags.append(TagDefinition('0008,0016', 'UI', 'SOPClassUID'))
        tags.append(TagDefinition('0008,0018', 'UI', 'SOPInstanceUID'))
        tags.append(TagDefinition('0008,0020', 'DA', 'StudyDate'))
        tags.append(TagDefinition('0008,0021', 'DA', 'SeriesDate'))
        tags.append(TagDefinition('0008,0060', 'CS', 'Modality'))
        tags.append(TagDefinition('0008,0070', 'LO', 'Manufacturer'))
        tags.append(TagDefinition('0008,1030', 'LO', 'StudyDescription'))
        tags.append(TagDefinition('0008,103E', 'LO', 'SeriesDescription'))

        tags.append(TagDefinition('0010,0010', 'PN', 'PatientName'))
        tags.append(TagDefinition('0010,0020', 'LO', 'PatientID'))
        tags.append(TagDefinition('0010,0030', 'DA', 'PatientsBirthDate'))

        tags.append(TagDefinition('0018,1250', 'SH', 'ReceiveCoilName'))

        tags.append(TagDefinition('0018,5100', 'CS', 'PatientPosition'))

        tags.append(TagDefinition('0020,000D', 'UI', 'StudyInstanceUID'))
        tags.append(TagDefinition('0020,000E', 'UI', 'SeriesInstanceUID'))
        tags.append(TagDefinition('0020,0010', 'SH', 'StudyID'))
        tags.append(TagDefinition('0020,0011', 'IS', 'SeriesNumber'))
        tags.append(TagDefinition('0020,0020', 'CS', 'PatientOrientation'))
        tags.append(TagDefinition('0020,0032', 'DS', 'ImagePositionPatient'))
        tags.append(TagDefinition('0020,0037', 'DS', 'ImageOrientationPatient'))
        tags.append(TagDefinition('0020,0052', 'UI', 'FrameOfReferenceUID'))
        tags.append(TagDefinition('0020,1002', 'IS', 'ImagesInAcquisition'))
        tags.append(TagDefinition('0020,1041', 'DS', 'SliceLocation'))

        tags.append(TagDefinition('0028,0002', 'US', 'SamplesPerPixel'))
        tags.append(TagDefinition('0028,0004', 'CS', 'PhotometricInterpretation'))
        tags.append(TagDefinition('0028,0006', 'US', 'PlanarConfiguration'))
        tags.append(TagDefinition('0028,0008', 'IS', 'NumberOfFrames'))
        tags.append(TagDefinition('0028,0010', 'US', 'Rows'))
        tags.append(TagDefinition('0028,0011', 'US', 'Columns'))
        tags.append(TagDefinition('0028,0030', 'DS', 'PixelSpacing'))
        tags.append(TagDefinition('0028,0100', 'US', 'BitsAllocated'))
        tags.append(TagDefinition('0028,0101', 'US', 'BitsStored'))
        tags.append(TagDefinition('0028,0102', 'US', 'HighBit'))
        tags.append(TagDefinition('0028,0103', 'US', 'PixelRepresentation'))
        tags.append(TagDefinition('0028,0120', 'SS', 'PixelPaddingValue'))
        tags.append(TagDefinition('0028,1050', 'DS', 'WindowCenter'))
        tags.append(TagDefinition('0028,1051', 'DS', 'WindowWidth'))
        tags.append(TagDefinition('0028,1052', 'DS', 'RescaleIntercept'))
        tags.append(TagDefinition('0028,1053', 'DS', 'RescaleSlope'))
        return tags



"""Custom classes for grouping DICOM files with dicomity"""

from dataclasses import dataclass, fields
from typing import Optional

from dataclasses_json import dataclass_json
from pydicom import FileDataset
from pydicom.multival import MultiValue
from pydicom.valuerep import PersonName


@dataclass_json
@dataclass
class GroupingMetadata:
    # pylint: disable=invalid-name
    """Store tags for an individual DICOM file used for grouping

    These are the tags used to group a set of Dicom Files into sets of coherent
    images
    """

    FileMetaInformationGroupLength: Optional[int] = None
    MediaStorageSOPClassUID: Optional[str] = None
    TransferSyntaxUID: Optional[str] = None
    ImplementationClassUID: Optional[str] = None
    ImageType: Optional[list[str]] = None

    SOPClassUID: Optional[str] = None
    SOPInstanceUID: Optional[str] = None
    StudyDate: Optional[str] = None
    SeriesDate: Optional[str] = None
    Modality: Optional[str] = None
    Manufacturer: Optional[str] = None
    StudyDescription: Optional[str] = None
    SeriesDescription: Optional[str] = None
    PatientName: Optional[PatientName] = None
    PatientID: Optional[str] = None
    PatientBirthDate: Optional[str] = None
    ReceiveCoilName: Optional[str] = None
    PatientPosition: Optional[str] = None
    StudyInstanceUID: Optional[str] = None
    SeriesInstanceUID: Optional[str] = None
    StudyID: Optional[str] = ''
    SeriesNumber: Optional[int] = ''
    PatientOrientation: Optional[str] = None
    ImagePositionPatient: Optional[list[float]] = None
    ImageOrientationPatient: Optional[list[float]] = None
    FrameOfReferenceUID: Optional[str] = None
    ImagesInAcquisition: Optional[int] = None
    SliceLocation: Optional[float] = None
    SamplesPerPixel: Optional[int] = None
    PhotometricInterpretation: Optional[str] = None
    PlanarConfiguration: Optional[int] = None
    NumberOfFrames: Optional[int] = None
    Rows: Optional[int] = None
    Columns: Optional[int] = None
    PixelSpacing: Optional[list[float]] = None
    BitsAllocated: Optional[int] = None
    BitsStored: Optional[int] = None
    HighBit: Optional[int] = None
    PixelRepresentation: Optional[int] = None
    PixelPaddingValue: Optional[int] = None
    WindowCenter: Optional[list[float]] = None
    WindowWidth: Optional[list[float]] = None
    RescaleIntercept: Optional[float] = None
    RescaleSlope: Optional[float] = None

    @staticmethod
    def groupingTags() -> list[str]:
        return [f.name for f in fields(GroupingMetadata)]

    @staticmethod
    def fromPyDicom(pydicom_meta: FileDataset):
        meta_data = GroupingMetadata()
        for tag in GroupingMetadata.groupingTags():
            meta_data._setOptionalTag(meta=pydicom_meta, tag=tag)
        meta_data.filename = pydicom_meta.filename
        return meta_data

    def _setOptionalTag(self, meta: FileDataset, tag: str):
        value = getattr(meta, tag, None)
        setattr(self, tag, GroupingMetadata.convert(value))

    @staticmethod
    def convert(value):
        if type(value) in [str, int, float, type(None)]:
            # Exact native types don't need to be converted
            return value
        if isinstance(value, int):
            # Convert pydicom int subclasses to int
            return int(value)
        elif isinstance(value, str):
            # Convert pydicom str subclasses to str
            return str(value)
        elif isinstance(value, float):
            # Convert pydicom float subclasses to float
            return float(value)
        elif isinstance(value, MultiValue):
            # Convert MultiValue to python list
            return [GroupingMetadata.convert(v) for v in value]
        elif isinstance(value, PersonName):
            # Convert special PatientName object
            return PatientName.from_pydicom(value)
        else:
            raise ValueError(f'GroupingMetadata does not currently support '
                             f'type {type(value).__name__} ')


class TagDefinition:
    """Used in creating a DicomDictionary"""

    def __init__(self, tag_string, vr_type, name):
        self.TagString = tag_string
        self.Tag = [int(tag_string[0:4], 16), int(tag_string[5:9], 16)]
        self.TagIndex = int(tag_string[0:4] + tag_string[5:9], 16)
        self.VRType = vr_type
        self.Name = name


@dataclass_json
@dataclass
class PatientName:
    FamilyName: str = ''
    GivenName: str = ''
    MiddleName: str = ''
    NamePrefix: str = ''
    NameSuffix: str = ''

    @staticmethod
    def from_pydicom(person_name: PersonName):
        return PatientName(
            FamilyName=person_name.family_name,
            GivenName=person_name.given_name,
            MiddleName=person_name.middle_name,
            NamePrefix=person_name.name_prefix,
            NameSuffix=person_name.name_suffix,
        )

    @property
    def visible_name(self) -> str:
        return ' '.join([self.NamePrefix, self.GivenName, self.MiddleName,
                         self.FamilyName, self.NameSuffix])



"""Utility functions for dicomity"""

import numpy as np
from natsort import natsorted

from dicomity.types import GroupingMetadata, PatientName


def compare_inexact(
        field_name: str,
        this_metadata,
        other_metadata,
        tolerance
):
    """Compare numerical attributes with a tolerance

    The attributes do not have to exist, but they must both exist or both not
    exist for a True match

    Args:
        field_name:
        this_metadata:
        other_metadata:
        tolerance:

    Returns:

    """

    # If this field does not exist in either metadata, then return a
    # true match
    if not hasattr(this_metadata, field_name) and not \
            hasattr(other_metadata, field_name):
        return True

    # If this field exists in one but not the other, return a false
    # match
    if hasattr(this_metadata, field_name) != \
            hasattr(other_metadata, field_name):
        return False

    # Get the values of this field
    field_this = getattr(this_metadata, field_name)
    field_other = getattr(other_metadata, field_name)

    # If the field values are of a different type, return a false match
    if not isinstance(field_this, type(field_other)):
        return False

    # Inexact numeric match
    return np.amax(np.abs(np.subtract(field_this, field_other))) < tolerance


# pylint: disable-next=too-many-return-statements
def compare_main_tags(
        this_metadata: GroupingMetadata,
        other_metadata: GroupingMetadata
):
    """

    Args:
        this_metadata: Metadata for first file
        other_metadata:  Metadata for second file

    Returns:
        True if the fields match. False if any fields no not match. False if
            any field exists in only one of the metadata
    """

    # Check for exact matches in certain fields
    fields_to_compare = [
        'PatientName', 'PatientID', 'PatientBirthDate', 'StudyInstanceUID',
        'SeriesInstanceUID', 'StudyID', 'StudyDescription', 'SeriesNumber',
        'SeriesDescription', 'StudyDate', 'SeriesDate', 'Rows', 'Columns',
        'PixelSpacing', 'PatientPosition', 'FrameOfReferenceUID', 'Modality',
        'MediaStorageSOPClassUID', 'ImageType', 'SOPClassUID',
        'ImplementationClassUID', 'ImagesInAcquisition', 'SamplesPerPixel',
        'PhotometricInterpretation', 'BitsAllocated', 'BitsStored', 'HighBit',
        'PixelRepresentation'
    ]

    fields_to_compare = [field for field in fields_to_compare if
                         hasattr(this_metadata, field)]
    fields_to_compare_other = [field for field in fields_to_compare if
                               hasattr(other_metadata, field)]

    # Fields should exist in both metadata or neither
    if fields_to_compare != fields_to_compare_other:
        return False

    for field_name in fields_to_compare:
        if getattr(this_metadata, field_name) != \
                getattr(other_metadata, field_name):
            return False

    if not compare_inexact('ImageOrientationPatient', this_metadata,
                           other_metadata, 0.5):
        return False

    # Verify that the images contain the same tags relating to slice location
    if hasattr(this_metadata, 'SliceLocation') != \
            hasattr(other_metadata, 'SliceLocation'):
        return False

    if hasattr(this_metadata, 'ImagePositionPatient') != \
            hasattr(other_metadata, 'ImagePositionPatient'):
        return False

    if hasattr(this_metadata, 'ImagePositionPatient') != \
            hasattr(other_metadata, 'ImagePositionPatient'):
        return False

    # If the positions match exactly, then these are should not be in the
    # same group - they may be duplicates, or different time points
    if hasattr(this_metadata, 'ImagePositionPatient') and \
            hasattr(other_metadata, 'ImagePositionPatient') \
            and compare_inexact('ImagePositionPatient', this_metadata,
                                other_metadata, 0.0001):
        return False

    return True


def add_optional_field(text, struct_name, field_name, only_if_nonempty):
    """

    Args:
        text:
        struct_name:
        field_name:
        only_if_nonempty:

    Returns:

    """
    if text and only_if_nonempty:
        return text
    return ", ".join(filter(None,
                            [text, getattr(struct_name, field_name, None)]))


def patient_name_to_strings(patient_name: 'PatientName' or str):
    """Return string representations of patient name

    A Dicom PatientName may contain multiple fields. This function
    performs a simple combination of fields to produce a patient name
    string, and also a short version using the first field that is found

    Args:
        patient_name: Patient name as a str or PatientName

    Returns:
        Tuple containing
            Patient name as a str
            Short version of patient name as a str

    """
    if isinstance(patient_name, str):
        return patient_name, patient_name
    if not isinstance(patient_name, PatientName):
        return "", ""

    name = ''
    short_name = ''

    name = add_optional_field(name, patient_name, 'FamilyName', False)
    name = add_optional_field(name, patient_name, 'GivenName', False)
    name = add_optional_field(name, patient_name, 'MiddleName', False)
    name = add_optional_field(name, patient_name, 'NamePrefix', False)
    name = add_optional_field(name, patient_name, 'NameSuffix', False)

    short_name = add_optional_field(short_name, patient_name, 'FamilyName',
                                    True)
    short_name = add_optional_field(short_name, patient_name, 'GivenName',
                                    True)
    short_name = add_optional_field(short_name, patient_name, 'MiddleName',
                                    True)
    short_name = add_optional_field(short_name, patient_name, 'NamePrefix',
                                    True)
    short_name = add_optional_field(short_name, patient_name, 'NameSuffix',
                                    True)
    return name, short_name


def sort(array):
    """Sort array, return sorted values and sorted indices

    Args:
        array: array to sort

    Returns:
        Tuple of sorted values, sorted indices
    """
    array = np.array(array)
    sorted_indices = np.argsort(array)
    sorted_values = array[sorted_indices]
    return sorted_values, sorted_indices


def sort_filenames(original_filenames: list[str]):
    """Sorts a list of filenames, taking into account numbers"""
    return natsorted(original_filenames)



from dataclasses import dataclass
from os.path import join

from dataclasses_json import dataclass_json


@dataclass_json
@dataclass
class CoreFilename:
    """A structure for holding a file name and path"""

    file_path: str = None
    file_name: str = None

    def fullFile(self):
        return join(self.file_path, self.file_name)



"""Classes for grouping DICOM data using dicomity"""

from collections import UserList
from dataclasses import dataclass
from statistics import mode
from typing import Optional

import math
import numpy as np

from pyreporting.reporting import get_reporting
from dicomity.util import compare_main_tags, sort

from dicomity.types import GroupingMetadata


@dataclass
class DicomItem:
    """Stores data about a Dicom file"""

    filename: str
    metadata: GroupingMetadata


class DicomStack(UserList):
    """Store a set of Dicom metadata structures, corresponding
    to a coherent sequence of images

    Created by the DicomGrouper class, which separates and groups images into
    coherent sequences according to their metadata.
    """

    def add(self, filename: str, metadata: GroupingMetadata):
        """Add a new item to the DicomStack

        Args:
            filename: Dicom file name
            metadata: Dicom metadata
        """
        self.data.append(DicomItem(filename=filename, metadata=metadata))

    def matches(self, other_metadata: GroupingMetadata):
        """Determine if the DICOM file with the specified metadata belongs in
        this group

        Args:
            other_metadata: other file metadata to compare against this one

        Returns:
            True if the metadata should be part of this group
        """

        if len(self.data) < 1:
            raise ValueError("There are no images in this DicomStack")

        additional_image = self.data[1].metadata if len(self.data) > 1 else None

        return are_images_groupable(
            self.data[0].metadata,
            other_metadata,
            additional_image
        )

    def sort_and_get_parameters(self):
        """Sorts the images according to slice location, and computes values for
        slice thickness and global origin"""
        sorted_indices, slice_thickness, global_origin_mm, sorted_positions = \
            self.sort_images_by_location()
        if len(self.data) > 1:
            if sorted_indices is None:
                get_reporting().warning(
                    identifier='DicomStack:UnableToSortFiles',
                    message='The images in this series may appear in the wrong '
                            'order because I was unable to determine the '
                            'correct ordering')
            else:
                self.data = [self.data[i] for i in sorted_indices]
        return slice_thickness, global_origin_mm, sorted_positions

    def sort_images_by_location(self):
        """Sort a series of Dicom images by their slice locations and calculate
        additional image parameters

        Returns:
            Tuple of:

            sorted_indices - The indices of the metadata structures in
                metadata_grouping, ordered by slice location

            slice_thickness - A typical slice thickness for this group of
                images

            global_origin_mm - The coordinates of the first voxel in the image
                volume

            sorted_positions - The ImagePositionPatient for
                each slice in the sorted

        """

        reporting = get_reporting()
        reporting.show_progress('Ordering images')
        reporting.update_progress_value(0)

        # Determine if ImagePositionPatient and SliceLocation tags exist.
        # It is sufficient to only check the metadata for one image, assuming
        # that the metadata was grouped prior to calling this function. The
        # grouping ensures that within a group, each of these tags exist for
        # either all or no images.
        representative_metadata = self.data[0].metadata
        if hasattr(representative_metadata, 'ImagePositionPatient'):
            image_positions_patient = [item.metadata.ImagePositionPatient
                                       for item in self.data]
        else:
            image_positions_patient = None

        if hasattr(representative_metadata, 'SliceLocation'):
            slice_locations = [item.metadata.SliceLocation
                               for item in self.data]
        else:
            slice_locations = None

        if hasattr(representative_metadata, 'InstanceNumber'):
            instance_numbers = [item.metadata.InstanceNumber
                                for item in self.data]
        else:
            instance_numbers = None

        # Try to calculate slice locations from the ImagePositionPatient tags
        if hasattr(representative_metadata, 'ImagePositionPatient'):

            # The first slice in the sequence may not be the actual first slice
            # in the image, but this does not matter
            first_slice_position = np.array(image_positions_patient[0])
            offset_positions = [np.array(item.metadata.ImagePositionPatient) -
                                first_slice_position for item in self.data]

            # Work out the relative direction of each slice relative to the
            # 'first' slice
            orientation_1 = np.array(
                representative_metadata.ImageOrientationPatient[0:3])
            orientation_2 = np.array(
                representative_metadata.ImageOrientationPatient[3:6])
            normal_vector = np.cross(orientation_1, orientation_2)
            # directions = [np.sign(np.dot(normal_vector, offset)) for offset in
            #                       offset_positions]

            # We compute the displacements of the image position, which assumes
            # the images form a cuboid stack. Really to compute the slice
            # spacing we should compute the distance of the image position to
            # the plane formed by the image orientation vectors
            image_slice_locations = [np.sign(np.dot(normal_vector, offset)) *
                                     math.sqrt(offset[0] ** 2 + offset[1] ** 2 +
                                               offset[2] ** 2) for offset
                                     in offset_positions]
            # image_slice_locations = directions.*sqrt(
            # offset_positions(:, 1).^2 + offset_positions(:, 2).^2 +
            # offset_positions(:, 3).^2)

            sorted_slice_locations, sorted_indices = sort(image_slice_locations)
            global_origin_mm = [
                min(p[0] for p in image_positions_patient),
                min(p[1] for p in image_positions_patient),
                min(p[2] for p in image_positions_patient)
            ]
            slice_thicknesses = [abs(sorted_slice_locations[i] -
                                     sorted_slice_locations[i - 1]) for i in
                                 range(1, len(sorted_slice_locations))]

        # If this tag is not present, we try the SliceLocation tags
        elif slice_locations:
            sorted_slice_locations, sorted_indices = sort(slice_locations)
            global_origin_mm = None
            slice_thicknesses = [abs(sorted_slice_locations[i] -
                                     sorted_slice_locations[i - 1]) for i in
                                 range(1, len(sorted_slice_locations))]

        # In the absense of the above tags, we sort by the instance number
        # (slice number). Ths is less reliable
        elif instance_numbers:
            _, sorted_indices = sort(instance_numbers)
            global_origin_mm = None
            slice_thicknesses = None

        # Otherwise, we set everything to empty
        else:
            sorted_indices = None
            global_origin_mm = None
            slice_thicknesses = None

        sorted_positions = None
        if image_positions_patient:
            sorted_positions = [image_positions_patient[i] for i in
                                sorted_indices]

        # Remove any zero thicknesses, which may indicate multiple slices at the
        # same position
        if slice_thicknesses and any(thickness < 0.01 for thickness in
                                     slice_thicknesses):
            reporting.warning(
                identifier='sort_images_by_location:ZeroSliceThickness',
                message='This image contains more than one image at the same slice '
                'position')
            slice_thicknesses = [thickness for thickness in slice_thicknesses
                                 if thickness > 0.01]

        # If we have no non-zero slice thicknesses (including the case where we
        # only have a single slice) then try and use the SliceThickness tag if
        # it exists. Otherwise, we just set to empty to indicate that we cannot
        # determine this
        if not slice_thicknesses:
            if hasattr(self.data[0], 'SliceThickness'):
                slice_thickness = self.data[0].SliceThickness
            else:
                slice_thickness = None
        else:
            slice_thickness = mode(slice_thicknesses)
            if any(thickness - slice_thickness > 0.01 for thickness in
                   slice_thicknesses):
                reporting.warning(
                    identifier='sort_images_by_location:InconsistentSliceThickness',
                    message='Not all slices have the same thickness')

        reporting.complete_progress()
        return sorted_indices, slice_thickness, global_origin_mm, \
            sorted_positions


class DicomGrouper(UserList):
    """Separate a series of Dicom images into groups of coherent images

    DicomGrouper splits a series of Dicom images into 'groups' of images with
    similar orientations and image properties. For example, a series containing
    a scout image will typically be separated into a group containing the scan
    images and another group containing the scout image. Similarly, a
    localizer series containing images in multiple orientations will typically
    be separated into a separate group for each orientation.
    """

    def add_item(self, filename: str, metadata: GroupingMetadata):
        """Add new metadata for an image. If the metadata is coherent with an
        existing group, add the image to that group. Otherwise, create a new
        group.

        Args:
            filename: path to image
            metadata: image metadata
        """

        self.find_stack(metadata).add(filename=filename, metadata=metadata)

    def number_of_groups(self) -> int:
        """Return the number of groups in this data"""
        return len(self.data)

    def get_stack(self, index: int) -> DicomStack:
        """Return a particular DicomStack"""
        return self.data[index]

    def largest_stack(self) -> Optional[DicomStack]:
        """Return the stack with the greatest number of images"""

        if len(self.data) < 1:
            return None

        return max(self.data, key=lambda g: len(g))

    def find_stack(self, metadata: GroupingMetadata) -> DicomStack:
        """Find a stack to which the specified image metadata can be added

        Searches for a DicomStack to which the new image metadata could be
        added while forming a coherent image set. If no existing group exists,
        create and return a new empty group

        Args:
            metadata: image metadata

        Returns:
            The DicomStack to which the image metadata belongs, or a new group
        """
        for grouping in self.data:
            if grouping.matches(metadata):
                return grouping
        new_stack = DicomStack()
        self.data.append(new_stack)
        return new_stack


def are_images_groupable(
        this_metadata: GroupingMetadata,
        other_metadata: GroupingMetadata,
        additional_metadata: Optional[GroupingMetadata] = None
):
    """Determine whether two Dicom images form a coherent sequence

    Compares the metadata from two Dicom images. If the images are from the same
    patient, study and series, and have similar orientations, data types and
    image types, then they are considered to be part of a coherent sequence.

    Args:
        this_metadata: metadata for the image being tested
        other_metadata: metadata for existing image in group
        additional_metadata: metadata for additional image. This is only used
            for checking that the image locations form a coherent set. Set to
            None if there are currently no other images in the group

    Returns:
        True if the images form a coherent sequence
    """

    # Check that the main Dicom tags match
    match = compare_main_tags(this_metadata, other_metadata)

    # Check that image coordinates lie in a straight line
    if additional_metadata is not None:
        match = match and compare_main_tags(additional_metadata, other_metadata)
        match = match and are_image_locations_consistent(
            this_metadata, other_metadata, additional_metadata)
    return match


def are_image_locations_consistent(
        first_metadata: GroupingMetadata,
        second_metadata: GroupingMetadata,
        third_metadata: GroupingMetadata,
        tolerance_mm: float = 10):
    """Determine if three Dicom images are parallel and lie approximately on
    a straight line, i.e. they form a volume

    Args:
        first_metadata: metadata of image 1
        second_metadata: metadata of image 2
        third_metadata: metadata of image 3
        tolerance_mm: maximum permitted absolute difference between each
            component of the normalised vectors

    Returns:
        True if three images lie approximately on a straight line
        (determined by the coordinates in the ImagePositionPatient Dicom tags)
    """

    # If the ImagePositionPatient tag is not present, assume the images are
    # consistent
    if (not hasattr(first_metadata, 'ImagePositionPatient')) and \
            (not hasattr(second_metadata, 'ImagePositionPatient')) and \
            (not hasattr(third_metadata, 'ImagePositionPatient')):
        return True

    # First get the image position
    first_position = np.array(first_metadata.ImagePositionPatient)
    second_position = np.array(second_metadata.ImagePositionPatient)
    third_position = np.array(third_metadata.ImagePositionPatient)

    # Next, compute direction vectors between the points
    direction_vec_1 = second_position - first_position
    direction_vec_2 = third_position - first_position

    # Find a scaling between the direction vectors
    scale_index_1 = np.argmax(np.abs(direction_vec_1))
    max_1 = np.abs(direction_vec_1)[scale_index_1]
    scale_index_2 = np.argmax(np.abs(direction_vec_2))
    max_2 = np.abs(direction_vec_2)[scale_index_2]

    if max_1 > max_2:
        scale_1 = 1
        scale_2 = direction_vec_1[scale_index_2]/direction_vec_2[scale_index_2]
    else:
        scale_1 = direction_vec_2[scale_index_1]/direction_vec_1[scale_index_1]
        scale_2 = 1

    # Scale
    scaled_direction_vector_1 = direction_vec_1*scale_1
    scaled_direction_vector_2 = direction_vec_2*scale_2

    # Find the maximum absolute difference between the normalised vectors
    difference = np.abs(scaled_direction_vector_2 - scaled_direction_vector_1)
    max_difference = np.amax(difference)

    return max_difference <= tolerance_mm



"""Functions for loading an image volume from DICOM files"""

from os.path import join

import numpy as np

from dicomity.util import sort_filenames
from pyreporting.reporting import get_reporting
from dicomity.dictionary import DicomDictionary
from dicomity.dicom import is_dicom, read_grouping_metadata, read_dicom_image
from dicomity.group import DicomGrouper, DicomStack


def load_main_image_from_dicom_files(filenames: str or list[str]):
    """Loads a series of DICOM files into a coherent 3D volume

    Args:
        filenames: filepath or list of filepaths

    Returns:
        Tuple of: image, representative_metadata, slice_thickness,
            global_origin_mm, sorted_positions

        image: a numpy array containing the 3D volume
        representative_metadata: metadata from one slice of the main group
        slice_thickness: the computed distance between centrepoints of each
            slice
        global_origin_mm: The mm coordinates of the image origin
        sorted_positions: Patient positions for each slice in the sorted order
    """

    reporting = get_reporting()

    # A single filename can be specified as a string
    if isinstance(filenames, str):
        filenames = [filenames]

    # Load the metadata from the DICOM images, and group into coherent sequences
    file_grouper = load_metadata_from_dicom_files(filenames=filenames)

    if file_grouper.number_of_groups() < 1:
        return None, None, None, None, None

    # Warn the user if we found more than one group, since the others will not
    # be loaded into the image volume
    if file_grouper.number_of_groups() > 1:
        reporting.warning(
            identifier='load_main_image_from_dicom_files:MultipleGroupings',
            message='I have removed some images from this dataset because the '
                    'images did not form a coherent set. This may be due to '
                    'the presence of scout images or dose reports, or '
                    'localizer images in multiple orientations. I have formed '
                    'a volume form the largest coherent set of images in the '
                    'same orientation.')

    # Choose the group with the most images
    main_group = file_grouper.largest_stack()

    # Sort the images into the correct order
    slice_thickness, global_origin_mm, sorted_positions = \
        main_group.sort_and_get_parameters()

    # Obtain a representative set of metadata tags from the first image in the
    # sequence
    representative_metadata = main_group[0].metadata

    # Load the pixel data
    image = load_images_from_stack(stack=main_group)

    return image, representative_metadata, slice_thickness, \
        global_origin_mm, sorted_positions


def load_metadata_from_dicom_files(filenames: str or list[str]) -> DicomGrouper:
    """Load metadata from a series of DICOM files

    Args:
        filenames: filenames can be a string for a single filename, or an
            array of strings

    Returns:
        a DicomGrouper object containing the metadata grouped into coherent
            sequences of images
    """

    # Get the default pyreporting object
    reporting = get_reporting()

    # Show progress dialog
    reporting.show_progress('Reading image metadata')
    reporting.update_progress_value(0)

    # A single filename can be specified as a string
    if isinstance(filenames, str):
        filenames = [filenames]

    # Sort the filenames into numerical order. Normally, this ordering will be
    # overruled by the ImagePositionPatient or SliceLocation tags, but in the
    # absence of other information, the numerical slice ordering will be used.
    sorted_filenames = sort_filenames(filenames)
    num_slices = len(filenames)

    # The DicomGrouper performs the sorting of image metadata
    dicom_grouper = DicomGrouper()

    dictionary = DicomDictionary.essential_dictionary_without_pixel_data()

    file_index = 0
    for next_file in sorted_filenames:
        file_name = next_file

        combined_file_name = file_name
        if is_dicom(combined_file_name):
            dicom_grouper.add_item(
                filename=combined_file_name,
                metadata=read_grouping_metadata(combined_file_name))
        else:
            # If not a Dicom image, exclude it from the set and warn user
            reporting.warning(
                identifier='load_metadata_from_dicom_files:NotADicomFile',
                message=f'load_metadata_from_dicom_files: The file '
                f'{combined_file_name} is not a DICOM file and will be '
                f'removed from this series.')

        reporting.update_progress_value(round(100 * file_index / num_slices))

        file_index += 1

    reporting.complete_progress()
    return dicom_grouper


def load_images_from_stack(stack: DicomStack) -> np.array:
    """Load metadata from a series of DICOM files

    Args:
        stack: a DicomStack containing metadata

    Returns:
        a numpy array containing the image volume
    """

    # Get the default pyreporting object
    reporting = get_reporting()

    reporting.show_progress('Reading pixel data')
    reporting.update_progress_value(0)

    num_slices = len(stack)

    # Load image slice
    first_image_slice = read_dicom_image(file_name=stack[0].filename)
    if first_image_slice is None:
        return None

    # Pre-allocate image matrix
    size_i = stack[0].metadata.Rows
    size_j = stack[0].metadata.Columns
    size_k = num_slices
    samples_per_pixel = stack[0].metadata.SamplesPerPixel

    # Pre-allocate image matrix
    data_type = first_image_slice.dtype
    if data_type == np.char:
        reporting.info(
            identifier='load_images_from_stack:SettingDatatypeToInt8',
            message='Char datatype detected. Setting to int8')
        data_type = 'int8'
    image_size = [size_i, size_j, size_k, samples_per_pixel] if \
        samples_per_pixel > 1 else [size_i, size_j, size_k]
    raw_image = np.zeros(image_size, data_type)
    if samples_per_pixel > 1:
        raw_image[:, :, 0, :] = first_image_slice
    else:
        raw_image[:, :, 0] = first_image_slice

    for file_index in range(1, num_slices):
        next_slice = read_dicom_image(file_name=stack[file_index].filename)
        if samples_per_pixel > 1:
            raw_image[:, :, file_index, :] = next_slice
        else:
            raw_image[:, :, file_index] = next_slice
        reporting.update_progress_value(round(100 * file_index / num_slices))

    reporting.complete_progress()
    return raw_image
